/**********************************************************************
 *
 * ResultBox.java -- a list of phrases in the Phind java interface
 *
 * Copyright 1997-2000 Gordon W. Paynter
 * Copyright 2000 The New Zealand Digital Library Project
 *
 * A component of the Greenstone digital library software
 * from the New Zealand Digital Library Project at the
 * University of Waikato, New Zealand.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *********************************************************************/

/*********************************************************************

This class is used in the Phind java applet (Phind.java).

A ResultBox holds the results of a query to phindcgi.  They deal mostly
with the information content of the query, and have methods for parsing the
input into phrase and document items.  They have little do with display:
ResultBoxes are shown to the user through "ResultDisplay" panels, and are
drawn using "ResultCanvas" objects.

**********************************************************************/


import java.awt.Panel;
import java.awt.BorderLayout;
import java.awt.Scrollbar;

import java.awt.Label;
import java.awt.Event;

import java.net.*;
import java.applet.*;

import java.util.Vector;


public class ResultBox extends Panel {
    
    // Objects at a "higher" level than this one
    Phind phind;
    ResultDisplay display;

    // Surroundig objects
    ResultBox prev, next;

    // The components contained by this one.
    ResultCanvas c;
    ResultTitle t;
    Scrollbar s;
    Panel label;

    // Information about the query
    String searchKey, searchPhrase;
    String searchCollection;
    
    // Information about the result set
    int numberOfExpansions;
    int expansionsRetrieved;
    int numberOfDocuments;
    int documentsRetrieved;
    int nextPhraseBlock;
    int nextDocumentBlock;
    
    // just why did i add these variables?
    int mode;
    final int initMode = 0;
    public final int emptyMode = 1;
    final int loadingMode = 2;
    final int finishedMode = 3;

    String buffer;
    boolean finished;


    // Create a ResultBox
    // given details of the search that generated it.
    ResultBox(Phind p, String collect, String key, String phrase, ResultBox rb) {

	super();
	mode = initMode;

	phind = p;
	display = null;
	next = null;
	prev = rb;
	if (prev != null) prev.next = this;

	numberOfExpansions = -1;
	expansionsRetrieved = 0;
	numberOfDocuments = -1;
	documentsRetrieved = 0;
	nextPhraseBlock = 1;
	nextDocumentBlock = 1;

	searchKey = key;
	searchPhrase = phrase;
	searchCollection = collect;

	setLayout(new BorderLayout());


	s = new Scrollbar(Scrollbar.VERTICAL);
	disableScrollbar();
	add("East", s);

	c = new ResultCanvas(this);
	add("Center", c);

	t = new ResultTitle(this);
	add("North", t);

	buffer = "";
	finished = false;
	mode = emptyMode;
    }

    static String describeContents(String phrase, String c) {
	return( "\"" +phrase + "\" in " + c + ".");
    }

    String describeContents() {
	return( describeContents(searchPhrase, searchCollection) );
    }


    // Reset the contents of the box
    void resetBox( ) {
	buffer = "";
	finished = false;
	c.resetCanvas();
	disableScrollbar();
	mode = emptyMode;
    
	numberOfExpansions = -1;
	expansionsRetrieved = 0;
	numberOfDocuments = -1;
	documentsRetrieved = 0;
    }

    void setStatus( String status ) {
	phind.setStatus(status);
    }

    void disableScrollbar() {
	if (s.isEnabled()) {
	    s.setValues(0, 1, 0, 1);
	    s.setLineIncrement(1);
	    s.setPageIncrement(1);
	    s.disable();
	}
    }

    // Are there displays previous to and after this?
    public boolean prevBoxExists () {
	return (prev != null);
    }
    public boolean nextBoxExists () {
	return (next != null);
    }


    void lookupPhrase(String key, String phrase, int queryMode) {
	buffer = "";
	finished = false;
	phind.searchForPhrase(this, key, phrase, queryMode);
	t.repaint();
    }

    public boolean handleEvent(Event event) {
	//    System.out.println("event: " + event.toString());
	if ( (event.target == s) && 
	     ( (event.id == Event.SCROLL_ABSOLUTE) ||
	       (event.id == Event.SCROLL_LINE_DOWN) ||
	       (event.id == Event.SCROLL_LINE_UP) ||
	       (event.id == Event.SCROLL_PAGE_DOWN) ||
	       (event.id == Event.SCROLL_PAGE_UP) ) ) {
	    c.repaint();
	    return true;
	} else {
	    return super.handleEvent(event);
	}
    }    


    // a new section of the byte stream is read
    void parseBytes(byte[] bytes) {
	mode = loadingMode;

	// update the buffer that holds the bytes read
	String input;
	try {
	    input = new String(bytes, new String("Latin1"));
	} catch (Exception e) {
	    System.err.println(e);
	    input = "";
	}
	buffer = buffer.concat(input);

	// if buffer ends with "</phinddata>" then we are finished reading
	finished = buffer.endsWith("</phinddata>\n");

	// Parse each line until there are no more lines to parse
	while(parseLine());

	if (finished) {
	    c.updateMorePhrasesMarker();
	    c.updateMoreDocumentsMarker();
	} else {
	    // We haven't finished, give some feedback about progress
	    setStatus("Connected to host, " + c.numberOfPhrases + " results read");
	}
    }

    // Parse a line of XML
    //
    // This isn't necessarily real XML.  We assume every tag begins and
    // ends on one line.  We read one line out of the buffer at a time,
    // then "add" them to the box with the addLine method.
    //
    // Return true if there is a another line here to parse, otherwise false.

    boolean parseLine() {

	// System.out.println("parseLine " + finished + "(" + expansionsRetrieved + ")");
	String item;
	int eol = buffer.indexOf((int) '\n');

	// If there are no carriage returns, the line is incomplete
	if (eol == -1) {
	    return false;
	}

	// The <expansion> tag
	else if (buffer.startsWith("<expansion ")) {
	    item = buffer.substring(0, eol);
	    buffer = buffer.substring(eol + 1);
	    addExpansionTag(item);
	    return true;
	}
 
	// The <document> tag
	else if (buffer.startsWith("<document ")) {
	    item = buffer.substring(0, eol);
	    buffer = buffer.substring(eol + 1);
	    addDocumentTag(item);
	    return true;
	}
 
	// The <expansionlist> tag
	else if (buffer.startsWith("<expansionlist ")) {
	    item = buffer.substring(0, eol);
	    buffer = buffer.substring(eol + 1);
	    addExpansionListTag(item);
	    return true;
	}
 
	// The <documentlist> tag
	else if (buffer.startsWith("<documentlist ")) {
	    item = buffer.substring(0, eol);
	    buffer = buffer.substring(eol + 1);
	    addDocumentListTag(item);
	    return true;
	}
 
	// The <phinddata> tag introduces the dataset
	else if (buffer.startsWith("<phinddata ")) {

	}

	// The </phinddata> tag concludes the dataset
	else if (buffer.startsWith("</phinddata>")) {
	    mode = finishedMode;
	    return false;
	}
	
	// Ignore the piece we've read
	item = buffer.substring(0, eol);
	buffer = buffer.substring(eol + 1);
	return true;
    }
    


    // Add an expansion list tag
    //
    // Given a string containing an XML expansinlist tag of the form:
    //   <expansionlist length="5" start="0" end="5">
    //   <expansionlist length="134" start="0" end="10">
    // Read the expansion information from the tag
    //
    // Return true if successful, otherwise false.

    boolean addExpansionListTag( String line ) {

	// System.out.println( "addExpansionListTag: " + line);

	String lenStr = "";
	int length = 0;

	// Break the tag down into its component parts
	line = line.substring(15);
	int nextSplit;
       	while (line.length() > 0) {
	    line = line.trim();

	    // Read the overall length of the list
	    if (line.startsWith("length")) {
		line = line.substring(8);
		nextSplit = line.indexOf((int) '"');
		if (nextSplit >= 1) {
		    lenStr = line.substring(0, nextSplit);
		    length = Integer.valueOf(lenStr).intValue();
		    line = line.substring(nextSplit + 1);
		} else {
		    System.err.println("addExpansionListTag: error parsing: " + line);
 		}
	    } 

	    // Read and ignore some other tag
	    else {
		nextSplit = line.indexOf((int) ' ');
		if (nextSplit >= 1) {
		    line = line.substring(nextSplit + 1);
		} else {
		    line = "";
		}
	    }
	}

	// Add the length data to the ResultBox
	numberOfExpansions = length;
	return true;
    }


    // Add an expansion tag
    //
    // Given a string containing an XML expansin tag of the form:
    //   <expansion num="3" id="8421" text="PEOPLE and TREES" tf="3" df="3"/>
    //   <expansion num="4" id="8696" text="FOREST TREES" tf="3" df="3"/>
    // Create a new ResultItem for display
    //
    // Return true if successful, otherwise false.

    boolean addExpansionTag( String line ) {

	// System.out.println( "addExpansionTag: " + line + " (" + expansionsRetrieved + ")");

	String numStr = "", idStr = "", tfStr = "", dfStr = "", body = "";

	// Break the tag down into its component parts
	line = line.substring(11);
	int nextSplit;
       	while (line.length() > 0) {
	    line = line.trim();

	    // Read the expansion number
	    if (line.startsWith("num")) {
		line = line.substring(5);
		nextSplit = line.indexOf((int) '"');
		if (nextSplit >= 1) {
		    numStr = line.substring(0, nextSplit);
		    line = line.substring(nextSplit + 1);
		} else {
		    System.err.println("ResultBox addExpansionTag: error parsing: " + line);
 		}
	    } 

	    // Read the phrase ID
	    else if (line.startsWith("id")) {
		line = line.substring(4);
		nextSplit = line.indexOf((int) '"');
		if (nextSplit >= 1) {
		    idStr = line.substring(0, nextSplit);
		    line = line.substring(nextSplit + 1);
		} else {
		    System.err.println("ResultBox addExpansionTag: error parsing: " + line);
 		}
	    } 

	    // Read the text of the phrase
	    else if (line.startsWith("text")) {
		line = line.substring(6);
		nextSplit = line.indexOf((int) '"');
		if (nextSplit >= 1) {
		    body = line.substring(0, nextSplit);
		    line = line.substring(nextSplit + 1);
		} else {
		    System.err.println("ResultBox addExpansionTag: error parsing: " + line);
 		}
	    } 

	    // Read the frequency
	    else if (line.startsWith("tf")) {
		line = line.substring(4);
		nextSplit = line.indexOf((int) '"');
		if (nextSplit >= 1) {
		    tfStr = line.substring(0, nextSplit);
		    line = line.substring(nextSplit + 1);
		} else {
		    System.err.println("ResultBox addExpansionTag: error parsing: " + line);
 		}
	    } 

	    // Read the document frequency
	    else if (line.startsWith("df")) {
		line = line.substring(4);
		nextSplit = line.indexOf((int) '"');
		if (nextSplit >= 1) {
		    dfStr = line.substring(0, nextSplit);
		    line = line.substring(nextSplit + 1);
		} else {
		    System.err.println("ResultBox addExpansionTag: error parsing: " + line);
 		}
	    } 

	    // Read some other tag
	    else {
		nextSplit = line.indexOf((int) ' ');
		if (nextSplit >= 1) {
		    line = line.substring(nextSplit + 1);
		} else {
		    line = "";
		}
	    }
	}
	
	// Create a new ResultItem and add it to the display
	ResultItem ri = new ResultItem(idStr, tfStr, dfStr, "", body, "");
	
	if (c.addResultItem(ri)) {
	    expansionsRetrieved++;
	    return true;
	}
	
	return false;
    }



    // Add a document list tag
    //
    // Given a string containing an XML documentlist tag of the form:
    //   <documentlist length="797" start="0" end="10">
    // Read the expansion information from the tag
    //
    // Return true if successful, otherwise false.

    boolean addDocumentListTag( String line ) {

	// System.out.println( "addExpansionListTag: " + line);

	String lenStr = "";
	int length = 0;

	// Break the tag down into its component parts
	line = line.substring(13);
	int nextSplit;
       	while (line.length() > 0) {
	    line = line.trim();

	    // Read the overall length of the list
	    if (line.startsWith("length")) {
		line = line.substring(8);
		nextSplit = line.indexOf((int) '"');
		if (nextSplit >= 1) {
		    lenStr = line.substring(0, nextSplit);
		    length = Integer.valueOf(lenStr).intValue();
		    line = line.substring(nextSplit + 1);
		} else {
		    System.err.println("addExpansionListTag: error parsing: " + line);
 		}
	    } 

	    // Read and ignore some other tag
	    else {
		nextSplit = line.indexOf((int) ' ');
		if (nextSplit >= 1) {
		    line = line.substring(nextSplit + 1);
		} else {
		    line = "";
		}
	    }
	}

	// Add the length data to the ResultBox
	numberOfDocuments = length;
	return true;
    }


    // Add an document tag
    //
    // Given a string containing an XML document tag of the form:
    //   <document num="2" hash="HASH424e64b811fdad933be69c" freq="1" title="CONTENS"/>
    //   <document num="3" hash="HASH01f08efd8752ad54ca0a99cf" freq="1" title="Tree mixtures"/>
    //
    // Create a new ResultItem for display
    //
    // Return true if successful, otherwise false.

    boolean addDocumentTag( String line ) {

	// System.out.println( "addDocumentTag: " + line + " (" + documentsRetrieved + ")");

	String num = "", hash = "", freq = "", title = "";
	line = line.substring(9);

	// Break the tag down into its component parts
	int nextSplit;
       	while (line.length() > 0) {
	    line = line.trim();

	    // Read the expansion number
	    if (line.startsWith("num")) {
		line = line.substring(5);
		nextSplit = line.indexOf((int) '"');
		if (nextSplit >= 1) {
		    num = line.substring(0, nextSplit);
		    line = line.substring(nextSplit + 1);
		} else {
		    System.err.println("ResultBox addExpansionTag: error parsing: " + line);
 		}
	    } 

	    // Read the document OID hash
	    else if (line.startsWith("hash")) {
		line = line.substring(6);
		nextSplit = line.indexOf((int) '"');
		if (nextSplit >= 1) {
		    hash = line.substring(0, nextSplit);
		    line = line.substring(nextSplit + 1);
		} else {
		    System.err.println("ResultBox addExpansionTag: error parsing: " + line);
 		}
	    } 

	    // Read the document frequency
	    else if (line.startsWith("freq")) {
		line = line.substring(6);
		nextSplit = line.indexOf((int) '"');
		if (nextSplit >= 1) {
		    freq = line.substring(0, nextSplit);
		    line = line.substring(nextSplit + 1);
		} else {
		    System.err.println("ResultBox addExpansionTag: error parsing: " + line);
 		}
	    } 

	    // Read the title
	    else if (line.startsWith("title")) {
		line = line.substring(7);
		nextSplit = line.indexOf((int) '"');
		if (nextSplit >= 1) {
		    title = line.substring(0, nextSplit);
		    line = line.substring(nextSplit + 1);
		} else {
		    System.err.println("ResultBox addExpansionTag: error parsing: " + line);
 		}
	    } 

	    // Read some other tag or close the string off
	    else {
		nextSplit = line.indexOf((int) ' ');
		if (nextSplit >= 1) {
		    line = line.substring(nextSplit + 1);
		} else {
		    line = "";
		}
	    }
	}
	
	// Create a new ResultItem and add it to the display
	ResultItem ri = new ResultItem(hash, title, freq);

	if (c.addResultItem(ri)) {
	    documentsRetrieved++;
	    return true;
	}
	
	return false;
    }
}




